﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using System;
using Firebase.Auth;
using Amanotes.TripleSDK.Core;
using AppleAuth;
using AppleAuth.Enums;
using AppleAuth.Extensions;
using AppleAuth.Interfaces;
using AppleAuth.Native;
using System.Security.Cryptography;
using System.Text;

namespace Amanotes.TripleSDK.Firebase
{
    public class AppleProvider : IAuthProvider
    {        
        public event Action<IAuthProvider> OnLinkSuccess;
        public event Action<IAuthProvider, string> OnLinkFail;
        public event Action<IAuthProvider> OnSignInSuccess;
        public event Action<IAuthProvider, string> OnSignInFail;

        public Provider providerType { get { return Provider.Apple; } }
        public string ProviderID { get { return "apple.com"; } }

        FirebaseAuthService auth;
        IAppleAuthManager _appleAuthManager;        
        string _rawNonce;
        const string AppleUserIdKey = "AppleUserId";


        public AppleProvider()
        {
            // If the current platform is supported
            if (AppleAuthManager.IsCurrentPlatformSupported)
            {
                // Creates a default JSON deserializer, to transform JSON Native responses to C# instances
                var deserializer = new PayloadDeserializer();
                // Creates an Apple Authentication manager with the deserializer
                _appleAuthManager = new AppleAuthManager(deserializer);

                // If at any point we receive a credentials revoked notification, we delete the stored User ID, and go back to login
                _appleAuthManager.SetCredentialsRevokedCallback(result =>
                {
                    Debug.Log("Received revoked callback " + result);
                    PlayerPrefs.DeleteKey(AppleUserIdKey);
                });

                // If we have an Apple User Id available, get the credential status for it
                if (PlayerPrefs.HasKey(AppleUserIdKey))
                {
                    var storedAppleUserId = PlayerPrefs.GetString(AppleUserIdKey);
                    CheckCredentialStatusForUserId(storedAppleUserId);
                }
            }
            auth = FirebaseAuthService.Instance;
        }

        public void SignInAndReplaceData()
        {
            if (AppleAuthManager.IsCurrentPlatformSupported)
            {
                AttemptQuickLogin(false);
            }
            else
            {
                Debug.LogWarning("Unsupported platform.");
                OnSignInFail?.Invoke(this, "Unsupported platform.");
            }
        }

        public void Link()
        {
            if (AppleAuthManager.IsCurrentPlatformSupported)
            {
                AttemptQuickLogin(true);
            }
            else
            {
                Debug.LogWarning("Unsupported platform.");
                OnSignInFail?.Invoke(this, "Unsupported platform.");
            }
        }

        public void SignOut() {   }

        public void Update()
        {
            // Updates the AppleAuthManager instance to execute
            // pending callbacks inside Unity's execution loop
            if (_appleAuthManager != null)
            {
                _appleAuthManager.Update();
            }
        }

        void SignInAndReplaceCallback(IAppleIDCredential appleIDCredential)
        {
            //Todo: get current AMA data

            var identityToken = Encoding.UTF8.GetString(appleIDCredential.IdentityToken);
            var authorizationCode = Encoding.UTF8.GetString(appleIDCredential.AuthorizationCode);
            var credential = OAuthProvider.GetCredential("apple.com", identityToken, _rawNonce, authorizationCode);
            auth.firebaseAuth.SignInWithCredentialAsync(credential).ContinueWith(task => {
                if (task.IsCanceled)
                {
                    Debug.LogWarning("SignInWithApple was canceled.");
                    OnSignInFail?.Invoke(this, "Login was canceled.");
                    return;
                }

                if (task.IsFaulted)
                {
                    Debug.LogError("SignInWithApple encountered an error: " + task.Exception);
                    OnSignInFail?.Invoke(this, "An account already exists with the same email address but different sign-in credentials. Sign in using a provider associated with this email address.");
                    return;
                }

                if (task.IsCompleted)
                {
                    Debug.Log("SignInWithApple completed.");
                    OnSignInSuccess?.Invoke(this);
                    //Todo: merge AMA data

                    return;
                }

                Debug.LogError("SignInWithApple does not complete. Unknown reason.");
                OnSignInFail?.Invoke(this, "Unknown error.");
            });
        }

        void LinkCallback(IAppleIDCredential appleIDCredential)
        {
            var identityToken = Encoding.UTF8.GetString(appleIDCredential.IdentityToken);
            var authorizationCode = Encoding.UTF8.GetString(appleIDCredential.AuthorizationCode);
            var credential = OAuthProvider.GetCredential("apple.com", identityToken, _rawNonce, authorizationCode);
            auth.firebaseAuth.CurrentUser.LinkWithCredentialAsync(credential).ContinueWith(task =>
            {
                if (task.IsCanceled)
                {
                    Debug.LogWarning("LinkWithApple canceled.");
                    OnSignInFail?.Invoke(this, "Login was cancelled.");
                    return;
                }

                if (task.IsFaulted)
                {
                    Debug.LogError("LinkWithApple encountered an error: " + task.Exception);
                    string msg = "switch account";
                    if (auth.ProviderIds.Count == 0)
                        msg = "login"; //First time link => Sign in                    
                    OnLinkFail?.Invoke(this, $"This Apple is already registered. Do you want to {msg}?");
                    return;
                }

                if (task.IsCompleted)
                {
                    Debug.Log("LinkWithApple completed.");
                    OnLinkSuccess?.Invoke(this);
                    return;
                }

                Debug.LogError("LinkWithApple does not completed. Unknown reason.");
                OnSignInFail?.Invoke(this, "Unknown error.");
            });
        }

        private void AttemptQuickLogin(bool isLink)
        {
            _rawNonce = GenerateRandomString(32);
            var nonce = GenerateSHA256NonceFromRawNonce(_rawNonce);
            var quickLoginArgs = new AppleAuthQuickLoginArgs(nonce);

            // Quick login should succeed if the credential was authorized before and not revoked
            _appleAuthManager.QuickLogin(
                quickLoginArgs,
                credential =>
                {
                    // If it's an Apple credential, save the user ID, for later logins
                    PlayerPrefs.SetString(AppleUserIdKey, credential.User);
                    IAppleIDCredential appleIDCredential = credential as IAppleIDCredential;
                    if (appleIDCredential != null)
                    {
                        if (isLink)
                            LinkCallback(appleIDCredential);
                        else
                            SignInAndReplaceCallback(appleIDCredential);
                    }
                    else
                    {
                        Debug.LogError("SignInApple failed. AppleIDCredential is null");
                        OnSignInFail?.Invoke(this, "Sign in with Apple failed.");
                    }
                },
                error =>
                {
                    // If Quick Login fails, we should show the normal sign in with apple menu, to allow for a normal Sign In with apple
                    var authorizationErrorCode = error.GetAuthorizationErrorCode();
                    Debug.LogWarning("Quick Login Failed " + authorizationErrorCode.ToString() + " " + error.ToString());
                    SignInWithApple();
                });
        }

        private void SignInWithApple()
        {
            _rawNonce = GenerateRandomString(32);
            var nonce = GenerateSHA256NonceFromRawNonce(_rawNonce);
            var loginArgs = new AppleAuthLoginArgs(LoginOptions.IncludeEmail | LoginOptions.IncludeFullName, nonce);

            _appleAuthManager.LoginWithAppleId(
                loginArgs,
                credential =>
                {
                    // If a sign in with apple succeeds, we should have obtained the credential with the user id, name, and email, save it
                    PlayerPrefs.SetString(AppleUserIdKey, credential.User);
                    IAppleIDCredential appleIDCredential = credential as IAppleIDCredential;
                    if (appleIDCredential != null)
                    {                        
                        if (appleIDCredential.FullName != null)
                            PlayerPrefs.SetString("appleName", appleIDCredential.FullName.ToLocalizedString());                        
                        LinkCallback(appleIDCredential);
                    }                        
                    else
                    {
                        Debug.LogError("SignInApple failed. AppleIDCredential is null");
                        OnSignInFail?.Invoke(this, "Sign in with Apple failed.");
                    }    
                },
                error =>
                {
                    var authorizationErrorCode = error.GetAuthorizationErrorCode();
                    Debug.LogError("Sign in with Apple failed " + authorizationErrorCode.ToString() + " " + error.ToString());
                    OnSignInFail?.Invoke(this, "Sign in with Apple failed.");
                });
        }

        private void CheckCredentialStatusForUserId(string appleUserId)
        {
            // If there is an apple ID available, we should check the credential state
            _appleAuthManager.GetCredentialState(
                appleUserId,
                state =>
                {
                    switch (state)
                    {
                    // If it's authorized, login with that user id
                    case CredentialState.Authorized:
                            return;

                    // If it was revoked, or not found, we need a new sign in with apple attempt
                    // Discard previous apple user id
                    case CredentialState.Revoked:
                        case CredentialState.NotFound:
                            PlayerPrefs.DeleteKey(AppleUserIdKey);
                            return;
                    }
                },
                error =>
                {
                    var authorizationErrorCode = error.GetAuthorizationErrorCode();
                    Debug.LogWarning("Error while trying to get credential state " + authorizationErrorCode.ToString() + " " + error.ToString());
                    OnSignInFail?.Invoke(this, "Error while trying to get credential.");
                });
        }

        private string GenerateRandomString(int length)
        {
            if (length <= 0)
            {
                throw new Exception("Expected nonce to have positive length");
            }

            const string charset = "0123456789ABCDEFGHIJKLMNOPQRSTUVXYZabcdefghijklmnopqrstuvwxyz-._";
            var cryptographicallySecureRandomNumberGenerator = new RNGCryptoServiceProvider();
            var result = string.Empty;
            var remainingLength = length;

            var randomNumberHolder = new byte[1];
            while (remainingLength > 0)
            {
                var randomNumbers = new List<int>(16);
                for (var randomNumberCount = 0; randomNumberCount < 16; randomNumberCount++)
                {
                    cryptographicallySecureRandomNumberGenerator.GetBytes(randomNumberHolder);
                    randomNumbers.Add(randomNumberHolder[0]);
                }

                for (var randomNumberIndex = 0; randomNumberIndex < randomNumbers.Count; randomNumberIndex++)
                {
                    if (remainingLength == 0)
                    {
                        break;
                    }

                    var randomNumber = randomNumbers[randomNumberIndex];
                    if (randomNumber < charset.Length)
                    {
                        result += charset[randomNumber];
                        remainingLength--;
                    }
                }
            }

            return result;
        }

        private string GenerateSHA256NonceFromRawNonce(string rawNonce)
        {
            var sha = new SHA256Managed();
            var utf8RawNonce = Encoding.UTF8.GetBytes(rawNonce);
            var hash = sha.ComputeHash(utf8RawNonce);

            var result = string.Empty;
            for (var i = 0; i < hash.Length; i++)
            {
                result += hash[i].ToString("x2");
            }

            return result;
        }
    }    
}
