﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using System;
using Facebook.Unity;
using Firebase.Auth;
using Amanotes.TripleSDK.Core;

namespace Amanotes.TripleSDK.Firebase
{
    public class FacebookProvider : IAuthProvider
    {        
        public event Action<IAuthProvider> OnLinkSuccess;
        public event Action<IAuthProvider, string> OnLinkFail;
        public event Action<IAuthProvider> OnSignInSuccess;
        public event Action<IAuthProvider, string> OnSignInFail;

        public Provider providerType { get { return Provider.Facebook; } }
        public string ProviderID { get { return "facebook.com"; } }

        FirebaseAuthService auth;

        public FacebookProvider()
        {
            if (!FB.IsInitialized)
            {
                // Initialize the Facebook SDK
                FB.Init(InitCallback, OnHideUnity);
            }
            else
            {
                // Already initialized, signal an app activation App Event
                FB.ActivateApp();
            }
            auth = FirebaseAuthService.Instance;
        }

        public void SignInAndReplaceData()
        {
            if (FB.IsLoggedIn)
                SignInAndReplaceCallback(null);
            else
                LoginRequest(SignInAndReplaceCallback);
        }

        public void Link()
        {
            LoginRequest(LinkCallback);
        }

        public void SignOut()
        {
            if (FB.IsLoggedIn)
                FB.LogOut();
        }

        void InitCallback()
        {
            if (FB.IsInitialized)
            {
                // Signal an app activation App Event
                FB.ActivateApp();
            }
            else
            {
                Debug.Log("Failed to Initialize the Facebook SDK");
            }
        }

        void OnHideUnity(bool isGameShown)
        {
            if (!isGameShown)
            {
                // Pause the game - we will need to hide
                Time.timeScale = 0;
            }
            else
            {
                // Resume the game - we're getting focus again
                Time.timeScale = 1;
            }
        }

        void LoginRequest(FacebookDelegate<ILoginResult> callback)
        {
            List<string> permissions = new List<string>
            {
                "public_profile",
                "email"
            };
            FB.LogInWithReadPermissions(permissions, callback);
        }

        void SignInAndReplaceCallback(ILoginResult result)
        {
            if (FB.IsLoggedIn)
            {
                //Todo: get current AMA data
                
                var facebookToken = AccessToken.CurrentAccessToken.TokenString;
                Credential credential = FacebookAuthProvider.GetCredential(facebookToken);                
                auth.firebaseAuth.SignInWithCredentialAsync(credential).ContinueWith(task => {
                    if (task.IsCanceled)
                    {
                        Debug.LogWarning("SignInWithFB was canceled.");
                        OnSignInFail?.Invoke(this, "Login was canceled.");
                        return;
                    }

                    if (task.IsFaulted)
                    {
                        Debug.LogError("SignInWithFB encountered an error: " + task.Exception);
                        OnSignInFail?.Invoke(this, "An account already exists with the same email address but different sign-in credentials. Sign in using a provider associated with this email address.");
                        return;
                    }

                    if (task.IsCompleted)
                    {
                        Debug.Log("SignInWithFB completed.");
                        OnSignInSuccess?.Invoke(this);
                        //Todo: merge AMA data

                        return;
                    }

                    Debug.LogError("SignInWithFB does not complete. Unknown reason.");
                    OnSignInFail?.Invoke(this, "Unknown error.");

                });
            }
            else
            {
                Debug.LogWarning("User has not logged in yet.");
                OnSignInFail?.Invoke(this, "User has not logged in yet.");
            }
        }

        void LinkCallback(ILoginResult result)
        {
            if (FB.IsLoggedIn)
            {
                // AccessToken class will have session details
                var facebookToken = AccessToken.CurrentAccessToken.TokenString;                

                Credential credential = FacebookAuthProvider.GetCredential(facebookToken);
                auth.firebaseAuth.CurrentUser.LinkWithCredentialAsync(credential).ContinueWith(task =>
                {
                    if (task.IsCanceled)
                    {
                        Debug.LogWarning("LinkWithFB canceled.");
                        OnSignInFail?.Invoke(this, "Login was canceled.");
                        return;
                    }
                        
                    if (task.IsFaulted)
                    {
                        Debug.LogError("LinkWithFB encountered an error: " + task.Exception);
                        if (auth.ProviderIds.Count == 0)
                            SignInAndReplaceCallback(null); //First time link => Sign in
                        else
                            OnLinkFail?.Invoke(this, "This FB is already registered. Do you want to switch account?");
                        return;
                    }

                    if (task.IsCompleted)
                    {
                        Debug.Log("LinkWithFB completed.");
                        OnLinkSuccess?.Invoke(this);
                        return;
                    }
                        
                    Debug.LogError("LinkWithFB does not completed. Unknown reason.");
                    OnSignInFail?.Invoke(this, "Unknown error.");
                });
            }
            else
            {
                Debug.LogWarning("User has not logged in yet.");
                OnSignInFail?.Invoke(this, "User has not logged in yet.");
            }
        }

        public string GetAvatarString()
        {
            if (!FB.IsLoggedIn)
            {
                Debug.LogError("Facebook is not logged in");
                return "";
            }

            AccessToken accessToken = AccessToken.CurrentAccessToken;
            return $"https://graph.facebook.com/v8.0/{accessToken.UserId}/picture?width=400&access_token={accessToken.TokenString}";            
        }
    }    
}
