﻿using System.Collections;
using System.Collections.Generic;
using System.Threading.Tasks;
using UnityEngine;
using Amanotes.TripleSDK.Core;
using Firebase.Auth;
using System;
using Google;

namespace Amanotes.TripleSDK.Firebase
{
    public class GoogleProvider : IAuthProvider
    {
        public event Action<IAuthProvider> OnLinkSuccess;
        public event Action<IAuthProvider, string> OnLinkFail;
        public event Action<IAuthProvider> OnSignInSuccess;
        public event Action<IAuthProvider, string> OnSignInFail;

        public Provider providerType { get { return Provider.Google; } }
        public string ProviderID { get { return "google.com"; } }

        FirebaseAuthService auth;

        public GoogleProvider()
        {
            GoogleSignIn.Configuration = new GoogleSignInConfiguration
            {
                WebClientId = AuthConfig.WEB_CLIENT_ID(),
                UseGameSignIn = false,
                RequestIdToken = true,
            };

            auth = FirebaseAuthService.Instance;
        }

        public void SignInAndReplaceData()
        {
            LoginRequest().ContinueWith(SignInAndReplaceCallback);
        }

        public void Link()
        {
            LoginRequest().ContinueWith(LinkCallback);            
        }

        public void SignOut()
        {
            GoogleSignIn.DefaultInstance.SignOut();            
        }

        Task<GoogleSignInUser> LoginRequest()
        {
            Debug.Log("Calling Google SignIn");
            return GoogleSignIn.DefaultInstance.SignIn();
        }

        void SignInAndReplaceCallback(Task<GoogleSignInUser> ggTask)
        {
            if (ggTask.IsFaulted)
            {
                Debug.LogError("Login GG gots error. " + ggTask.Exception);
                OnSignInFail?.Invoke(this, "Fail to login with Google");
                return;
            }

            if (ggTask.IsCanceled)
            {
                Debug.LogWarning("Login GG is canceled");
                OnSignInFail?.Invoke(this, "Login was canceled");
                return;
            }

            if (ggTask.IsCompleted)
            {
                //Todo: get current AMA data

                var ggToken = ggTask.Result.IdToken;
                Credential credential = GoogleAuthProvider.GetCredential(ggToken, null);
                auth.firebaseAuth.SignInWithCredentialAsync(credential).ContinueWith(task => {
                    if (task.IsCanceled)
                    {
                        Debug.LogWarning("SignInWithGG canceled.");
                        OnSignInFail?.Invoke(this, "Login was canceled");
                        return;
                    }

                    if (task.IsFaulted)
                    {
                        Debug.LogError("SignInWithGG encountered an error: " + task.Exception);
                        OnSignInFail?.Invoke(this, "An account already exists with the same email address but different sign-in credentials. Sign in using a provider associated with this email address.");
                        return;
                    }

                    if (task.IsCompleted)
                    {
                        Debug.Log("SignInWithGG completed.");
                        OnSignInSuccess?.Invoke(this);                  
                        //Todo: merge AMA data

                        return;
                    }

                    Debug.LogError("SignInWithGG does not complete. Unknown reason.");
                    OnSignInFail?.Invoke(this, "Unknown error.");
                });
                return;
            }
        }

        void LinkCallback(Task<GoogleSignInUser> ggTask)
        {
            if (ggTask.IsFaulted)
            {
                Debug.LogError("Login GG gots error. " + ggTask.Exception);
                OnSignInFail?.Invoke(this, "Fail to link with Google");
            }
            else if (ggTask.IsCanceled)
            {
                Debug.LogWarning("Login GG is canceled");
                OnSignInFail?.Invoke(this, "Cancel to link with Google");
            }
            else
            {
                var ggToken = ggTask.Result.IdToken;
                Credential credential = GoogleAuthProvider.GetCredential(ggToken, null);
                auth.firebaseAuth.CurrentUser.LinkWithCredentialAsync(credential).ContinueWith(task =>
                {
                    if (task.IsCanceled)
                    {
                        Debug.LogWarning("LinkWithGG was canceled.");
                        OnLinkFail?.Invoke(this, "Cancel to link with Google");
                        return;
                    }

                    if (task.IsFaulted)
                    {
                        Debug.LogError("LinkWithGG encountered an error: " + task.Exception);
                        if (auth.ProviderIds.Count == 0)
                        {
                            //First time link => Sign in
                            SignInAndReplaceCallback(ggTask);
                            return;
                        }                        
                        OnLinkFail?.Invoke(this, "This GG is already registered. Do you want to switch account?");
                        return;
                    }

                    if (task.IsCompleted)
                    {
                        Debug.Log("LinkWithGG completed.");
                        OnLinkSuccess?.Invoke(this);
                        return;
                    }

                    Debug.LogError("LinkWithGG does not complete. Unknown reason.");
                    OnSignInFail?.Invoke(this, "Unknown error.");
                });
            }
        }
    }
}