﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using Firebase.Auth;
using System;
using Amanotes.TripleSDK.Core;

namespace Amanotes.TripleSDK.Firebase
{
    public class PhoneProvider : IAuthProvider
    {
        public event Action<IAuthProvider> OnLinkSuccess;
        public event Action<IAuthProvider, string> OnLinkFail;
        public event Action<IAuthProvider> OnSignInSuccess;
        public event Action<IAuthProvider, string> OnSignInFail;
        public event Action<string> OnVerifyFail;

        public Provider providerType { get { return Provider.Phone; } }
        public string ProviderID { get { return "phone"; } }

        public string phoneNumber;
        public string verificationCode;

        public enum Purpose
        {
            Link,
            SignIn
        }

        PhoneAuthProvider provider;
        string verificationId;
        Purpose curPurpose;
        FirebaseAuthService auth;

        public PhoneProvider()
        {
            provider = PhoneAuthProvider.GetInstance(FirebaseAuth.DefaultInstance);
            auth = FirebaseAuthService.Instance;
        }

        public void SignInAndReplaceData() { }

        public void Link()
        {
            if (verificationId != "" && verificationCode != "")
            {
                Credential credential = provider.GetCredential(verificationId, verificationCode);
                if (!credential.IsValid())
                {
                    OnVerifyFail?.Invoke("Wrong verification code");
                    return;
                }

                switch (curPurpose)
                {
                    case Purpose.Link:
                        LinkCallback(credential);
                        break;
                    case Purpose.SignIn:
                        SignInAndReplaceData(credential);
                        break;
                }
            }
            else
            {
                OnVerifyFail?.Invoke("Not yet provide verification code");
            }
        }

        public void SignOut()
        {
            verificationId = string.Empty;
            verificationCode = string.Empty;
            phoneNumber = string.Empty;
        }

        public void CheckPhoneAuth()
        {
            //Todo: check phone auth exists
            VerifyPhoneNumber(Purpose.Link);
        }

        void CheckPhoneAuthCallback(bool authPhoneExist)
        {
            if (authPhoneExist)
                if (auth.ProviderIds.Count == 0)
                    VerifyPhoneNumber(Purpose.SignIn);
                else
                    Dialogue.Instance.Show("Phone number is already registered. Do you want to switch account?",
                        onYes: () => VerifyPhoneNumber(Purpose.SignIn));
            else
                VerifyPhoneNumber(Purpose.Link);
        }

        void VerifyPhoneNumber(Purpose purpose)
        {
            curPurpose = purpose;
            provider.VerifyPhoneNumber(phoneNumber, AuthConfig.PHONE_VERIFY_TIMEOUT, null,
                verificationCompleted: (credential) => {
                    Debug.Log("Phone: " + "auto-verify");
                    Loading.Instance.gameObject.SetActive(true);
                    switch (purpose)
                    {
                        case Purpose.Link:
                            LinkCallback(credential);
                            break;
                        case Purpose.SignIn:
                            SignInAndReplaceData(credential);
                            break;
                    }
                },
                verificationFailed: (error) => {
                // The verification code was not sent.
                Debug.Log("Phone: " + "code sent fail");
                    OnVerifyFail?.Invoke("Fail to send verification code");
                },
                codeSent: (id, token) => {
                // Verification code was successfully sent via SMS.
                // `id` contains the verification id that will need to passed in with
                // the code from the user when calling GetCredential().
                // `token` can be used if the user requests the code be sent again, to
                // tie the two requests together.
                Debug.Log("Phone: " + "sent code");
                    verificationId = id;
                },
                codeAutoRetrievalTimeOut: (id) => {
                // Called when the auto-sms-retrieval has timed out, based on the given
                // timeout parameter.
                // `id` contains the verification id of the request that timed out.
                Debug.Log("Phone: " + "auto-verify fail");
                    verificationId = id;
                });
        }

        void SignInAndReplaceData(Credential credential)
        {
            //Todo: get current AMA data        

            auth.firebaseAuth.SignInWithCredentialAsync(credential).ContinueWith(task => {
                if (task.IsCanceled)
                {
                    Debug.LogError("SignInWithPhone was canceled.");
                    return;
                }

                if (task.IsFaulted)
                {
                    Debug.LogError("SignInWithPhone encountered an error: " + task.Exception);
                    OnVerifyFail?.Invoke(task.Exception.ToString());
                    return;
                }

                if (task.IsCompleted)
                {
                    Debug.LogFormat("SignInWithPhone completed.");
                    OnSignInSuccess?.Invoke(this);
                    //Todo: merge AMA data

                }

                Debug.LogError("SignInWithPhone does not complete. Unknown reason.");

            });
        }

        void LinkCallback(Credential credential)
        {
            auth.firebaseAuth.CurrentUser.LinkWithCredentialAsync(credential).ContinueWith(task =>
            {
                if (task.IsCanceled)
                {
                    Debug.LogError("LinkWithPhone was canceled.");
                    return;
                }
                if (task.IsFaulted)
                {
                    Debug.LogError("LinkWithPhone encountered an error: " + task.Exception);
                    OnLinkFail?.Invoke(this, "This phone number is already registered. Please use another one.");
                    return;
                }

                OnLinkSuccess?.Invoke(this);
            });
        }
    }
}