﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using OFB = Firebase;
using Firebase.Extensions;
using Firebase.Auth;
using System;
using System.Threading.Tasks;
using System.IO;
using Amanotes.TripleSDK.Core;
using Amanotes.TripleSDK.Core.Analytics;
using UnityEngine.SceneManagement;

namespace Amanotes.TripleSDK.Firebase
{
    [InjectService(Package = "AmaFirebase_Dotnet4")]
    [StartupService]
    public class FirebaseAuthService : FirebaseService
    {        
        public static FirebaseAuthService Instance;        

        private bool _enable = false;
        public override bool Enable
        {
            get => _enable;
            set
            {
                _enable = value;
                if (_enable)
                {
                    ServiceStatus = ServiceStatus.Available;
                    Bootstrap.Instance.GetRequiredService<DoOnMainThread>().QueueOnMainThread(() =>
                    {
                        InitializeFirebase();                        
                    });
                }
            }
        }

        public event Action OnSignInSuccess;
        public event Action OnLinkSuccess;
        public event Action OnLinkFail;
        
        public FirebaseAuth firebaseAuth { get; private set; }
        OFB.DependencyStatus dependencyStatus = OFB.DependencyStatus.UnavailableOther;

        public bool Available { private set; get; }
        public User User { private set; get; }
        public Texture2D UserImg { private set; get; }
        public List<Provider> ProviderIds { private set; get; }
        public AuthCountlyService Countly { get; private set; }

        string localSharedFile;
        string amaToken;
        string firebaseToken;
        string firebaseId;
        //FacebookProvider _FBProvider;
        //GoogleProvider _GGProvider;
        AppleProvider _APProvider;
        Dictionary<Provider, IAuthProvider> authProviders = new Dictionary<Provider, IAuthProvider>();

        #region MonoBehaviour
        private void Awake()
        {
            if (Instance == null)
            {
                Instance = this;
                DontDestroyOnLoad(gameObject);
            }
            else if (Instance != this)
            {
                Destroy(gameObject);
                return;
            }

            localSharedFile = Utils.GetSharedFolderPath(AuthConfig.APP_GROUP_ID()) + "amanotes.txt";
        }

        private void FirebaseAuthService_AMASignInSuccess()
        {
            Available = true;
        }

        // Start is called before the first frame update
        public override void Startup()
        {
            base.Startup();
        }

        private void Update()
        {
            // Updates the AppleAuthManager instance to execute
            // pending callbacks inside Unity's execution loop
            if (_APProvider != null)
                _APProvider.Update();
        }

        void OnDestroy()
        {
            if (firebaseAuth != null)
            {
                firebaseAuth.StateChanged -= AuthStateChanged;
                firebaseAuth = null;
            }
        }
        #endregion MonoBehaviour

        #region Firebase
        void CheckDependencies()
        {
            OFB.FirebaseApp.CheckAndFixDependenciesAsync().ContinueWithOnMainThread(task => {
                dependencyStatus = task.Result;
                if (dependencyStatus == OFB.DependencyStatus.Available)
                {
                    InitializeFirebase();
                }
                else
                {
                    Debug.LogError(
                      "Could not resolve all Firebase dependencies: " + dependencyStatus);
                }
            });
        }

        // Handle initialization of the necessary firebase modules:
        protected void InitializeFirebase()
        {
            Debug.Log("Setting up Firebase Auth");
            OFB.AppOptions options = new OFB.AppOptions();
            options.ProjectId = AuthConfig.FIREBASE_AUTH_PROJECTID();
            options.ApiKey = AuthConfig.FIREBASE_AUTH_APIKEY();
            var app = OFB.FirebaseApp.Create(options, "AmaAuth");
            firebaseAuth = FirebaseAuth.GetAuth(app);            
            
            var fb = new FacebookProvider();
            authProviders.Add(fb.providerType, fb);
            var gg = new GoogleProvider();
            authProviders.Add(gg.providerType, gg);
            _APProvider = new AppleProvider();
            authProviders.Add(_APProvider.providerType, _APProvider);

            foreach (var p in authProviders.Values)
            {
                p.OnLinkSuccess += LinkSuccessHandler;
                p.OnLinkFail += LinkFailHandler;
                p.OnSignInFail += SignInFailHandler;
                p.OnSignInSuccess += SignInSuccessHandler;
            }

            OnSignInSuccess += FirebaseAuthService_AMASignInSuccess;
            
            firebaseAuth.StateChanged += AuthStateChanged;

            User = new User();
            Countly = gameObject.AddComponent<AuthCountlyService>();            
        }

        // Track state changes of the auth object.
        void AuthStateChanged(object sender, EventArgs eventArgs)
        {            
            if (firebaseAuth.CurrentUser != null && firebaseAuth.CurrentUser.UserId != firebaseId)
            {
                Debug.Log("AuthStateChanged Signed in " + firebaseAuth.CurrentUser.UserId);
                firebaseId = firebaseAuth.CurrentUser.UserId;

                GetFirebaseToken().ContinueWithOnMainThread(taskToken =>
                {
                    LoginData t = new LoginData
                    {
                        idToken = taskToken.Result,
                        providerId = firebaseAuth.CurrentUser.ProviderId.ToLower() == Provider.Firebase.ToString().ToLower() ? "anonymous" : firebaseAuth.CurrentUser.ProviderId,
                        gameId = AuthConfig.GAME_ID
                    };
                    string data = JsonUtility.ToJson(t);
                    WebRequestHelper.Post(AuthConfig.SERVER_URL() + AuthConfig.POST_LOGIN(), data).ContinueWithOnMainThread(SignInPOSTCallback);
                });
            }
            else
            {
                //Auto log in anonymous
                SignInAnonymouslyAsync();
            }
        }

        //Sign in callback
        void SignInPOSTCallback(Task<ResponseData> taskPost)
        {
            ResponseData res = taskPost.Result;
            if (res.error != null)
            {
                Debug.LogError("[SignInPOSTCallback] " + res.error.message);

                var context = Bootstrap.GetService<ULSContext>();
                context.LoginPlatform = firebaseAuth.CurrentUser.ProviderId.ToString();
                context.Screen = SceneManager.GetActiveScene().name;               
                context.AmaId = "";
                context.Result = false;
                Bootstrap.GetService<IULSAnalytics>().LogEventLogin();

                SignOut();
                return;
            }

            LogInResponseData loginInfo = JsonHelper.FromJsonToObject<LogInResponseData>(res.data.ToString());
            AmaTokenData amaTokenData = loginInfo.data;

            amaToken = amaTokenData.accessToken;
            PlayerPrefs.SetString(AuthConfig.PREF_AMA_ID, amaTokenData.amaId);

            //Debug.Log(string.Format(
            //    $"<b>AMA Id:</b>{amaTokenData.amaId}\n" +
            //    $"<b>Token:</b>{amaTokenData.accessToken}\n")
            //    );

            UpdateUserInfo(amaTokenData.user).ContinueWithOnMainThread(_ =>
            {
                OnSignInSuccess?.Invoke();

                var context = Bootstrap.GetService<ULSContext>();
                context.LoginPlatform = GetLoginPlatform();
                context.Screen = SceneManager.GetActiveScene().name;                
                context.AmaId = amaTokenData.amaId;
                context.Result = true;
                Bootstrap.GetService<IULSAnalytics>().LogEventLogin();
            });
        }

        // Fetch and display current user's auth token.
        public async Task<string> GetFirebaseToken()
        {
            if (firebaseAuth.CurrentUser == null)
            {
                Debug.Log("Not signed in, unable to get token.");
                //return Task<string>.Factory.StartNew(() => { return string.Empty; });
                firebaseToken = string.Empty;
            }
            else
                await firebaseAuth.CurrentUser.TokenAsync(false).ContinueWithOnMainThread(task => {
                    //Debug.Log("Firebase Token = " + task.Result);
                    firebaseToken = task.Result;
                });
            return firebaseToken;
        }
        #endregion Firebase

        public void LinkWith(Provider provider)
        {
#if UNITY_EDITOR
            DoesNotWorkOnEditor();
#else
            authProviders[provider].Link();
#endif
        }

        // Attempt to sign in anonymously.
        public Task<FirebaseUser> SignInAnonymouslyAsync()
        {
            Debug.Log("Attempting to sign anonymously...");
            return firebaseAuth.SignInAnonymouslyAsync();
        }

        public void SignOut()
        {
            Debug.Log("Signing out.");
            PlayerPrefs.SetString(AuthConfig.PREF_AMA_ID, "");
            foreach (var p in authProviders.Values)
            {
                p.SignOut();
            }
            firebaseAuth.SignOut();

            var context = Bootstrap.GetService<ULSContext>();
            context.AmaId = User.amaId;
            context.Result = true;
            Bootstrap.GetService<IULSAnalytics>().LogEventLogout();
        }

        // Display information about the currently logged in user.
        public async Task GetUserInfo()
        {
            if (firebaseAuth.CurrentUser == null || string.IsNullOrEmpty(amaToken))
            {
                Debug.Log("Not signed in, unable to get info.");
                return;
            }

            await WebRequestHelper.Get(AuthConfig.SERVER_URL() + AuthConfig.GET_USER(), amaToken).ContinueWithOnMainThread(async task =>
            {
                ResponseData res = task.Result;
                if (res.error != null)
                {
                    Debug.LogError("[GetUserInfo] " + res.error.message);
                    return;
                }

                await UpdateUserInfo(res.data.ToString());

                Bootstrap.Instance.GetRequiredService<IAnalytics>()?.LogEvent("firebase_get_user",
                    new Dictionary<string, object> { { "username", User.username } });
            });
        }

        public void ChangeUsername(string newName, Action<string> callback)
        {
            string oldName = User.username;
            UserPut userPut = new UserPut
            {
                username = newName
            };
            string data = JsonUtility.ToJson(userPut);
            WebRequestHelper.Put(AuthConfig.SERVER_URL() + AuthConfig.PUT_USERNAME(), amaToken, data).ContinueWithOnMainThread(taskPut =>
            {
                ResponseData res = taskPut.Result;
                if (res.error != null)
                {
                    Debug.LogWarning("[ChangeUsername] " + res.error.message);
                    callback(res.error.message);
                    return;
                }
                else
                {
                    User.username = newName;
                    callback("");
                }                    
            });
            
        }

        public string GetLocalSavedName()
        {
            //Check local data exist
            if (File.Exists(localSharedFile))
            {
                try
                {
                    string strData = File.ReadAllText(localSharedFile);
                    LocalData localData = JsonUtility.FromJson<LocalData>(strData);
                    return localData.fbName;
                }
                catch (Exception ex)
                {
                    Debug.LogWarning("Cannot load local data. " + ex.Message);
                    return string.Empty;
                }
            }
            else
            {
                return string.Empty;
            }
        }

        async Task UpdateUserInfo(string rawData)
        {            
            if (string.IsNullOrEmpty(rawData))
            {
                Debug.LogWarning("No user data");
            }
            else
            {
                GetUserResponseData userRes = JsonHelper.FromJsonToObject<GetUserResponseData>(rawData);              
                await UpdateUserInfo(userRes.data.user);
            }
        }

        async Task UpdateUserInfo(User u)
        {
            User = u;

            ProviderIds = new List<Provider>();
            if (User.isLoggedInFB)
                ProviderIds.Add(Provider.Facebook);
            if (User.isLoggedInGG)
                ProviderIds.Add(Provider.Google);
            if (User.isLoggedInApple)
            {
                ProviderIds.Add(Provider.Apple);
                User.appleName = PlayerPrefs.GetString("appleName");                
            }                
            if (User.isLoggedInPhone)
                ProviderIds.Add(Provider.Phone);

            //Download avatar
            string photoURL = User.photoURL;
            if (User.isLoggedInFB)
            {
                photoURL = ((FacebookProvider)authProviders[Provider.Facebook]).GetAvatarString();
            }
            if (!string.IsNullOrEmpty(photoURL))
                await WebRequestHelper.DownloadImage(photoURL).ContinueWithOnMainThread(task =>
                {
                    ResponseData res = task.Result;
                    if (res.error != null)
                    {
                        Debug.LogError("[UpdateUserInfo] " + res.error.message);
                    }
                    else
                    {                        
                        UserImg = (Texture2D)task.Result.data;                        
                    }
                });
        }

        void LinkSuccessHandler(IAuthProvider provider)
        {
            GetFirebaseToken().ContinueWithOnMainThread(taskFirebase =>
            {
                UserLink l = new UserLink
                {
                    idToken = firebaseToken,
                    providerId = provider.ProviderID
                };                
                string data = JsonUtility.ToJson(l);                
                WebRequestHelper.Post(AuthConfig.SERVER_URL() + AuthConfig.POST_LINK(), data).ContinueWithOnMainThread(task =>
                {
                    ResponseData res = task.Result;
                    if (res.error != null)
                    {
                        Debug.LogError("[LinkSuccessHandler] " + res.error.message);
                        OnLinkFail?.Invoke();
                        return;
                    }

                    Debug.Log("Track link account to server successfully");
                    UpdateUserInfo(res.data.ToString()).ContinueWithOnMainThread(_ =>
                    {
                        OnLinkSuccess?.Invoke();

                        var context = Bootstrap.GetService<ULSContext>();
                        context.PassportPlatform = provider.providerType.ToString().ToLower();
                        context.Result = true;
                        Bootstrap.GetService<IULSAnalytics>().LogEventLoginPassport();
                    });
                });
            });
        }

        void LinkFailHandler(IAuthProvider provider, string msg)
        {
            var context = Bootstrap.GetService<ULSContext>();
            context.PassportPlatform = provider.providerType.ToString().ToLower();
            context.Result = false;
            Bootstrap.GetService<IULSAnalytics>().LogEventLoginPassport();

            Dialogue.Instance.Show(msg, onYes: () => provider.SignInAndReplaceData());
        }

        void SignInSuccessHandler(IAuthProvider provider)
        {
            GetFirebaseToken().ContinueWithOnMainThread(taskToken =>
            {
                LoginData t = new LoginData
                {
                    idToken = taskToken.Result,
                    providerId = firebaseAuth.CurrentUser.ProviderId.ToLower() == Provider.Firebase.ToString().ToLower() ? "anonymous" : firebaseAuth.CurrentUser.ProviderId,
                    gameId = AuthConfig.GAME_ID
                };
                string data = JsonUtility.ToJson(t);
                WebRequestHelper.Post(AuthConfig.SERVER_URL() + AuthConfig.POST_LOGIN(), data).ContinueWithOnMainThread(SignInPOSTCallback);
            });
        }

        void SignInFailHandler(IAuthProvider provider, string msg)
        {
            var context = Bootstrap.GetService<ULSContext>();
            context.LoginPlatform = provider.providerType.ToString().ToLower();
            context.Screen = SceneManager.GetActiveScene().name;
            context.AmaId = "";
            context.Result = false;
            Bootstrap.GetService<IULSAnalytics>().LogEventLogin();

            Dialogue.Instance.Show(msg, onOK: OnLinkFail);
        }

        void DoesNotWorkOnEditor()
        {
            Debug.LogWarning("Does not work on editor");
            OnLinkFail?.Invoke();
        }

        string GetLoginPlatform()
        {
            if (User.isLoggedInFB)
                return "facebook";
            else if (User.isLoggedInGG)
                return "google";
            else if (User.isLoggedInApple)
                return "apple";
            else
                return "firebase";
        }
    }
}